@@ -1,8 +1,8 @@
-#ifndef DEFINES_H_INC
- #define DEFINES_H_INC
+#ifndef DEFINES_H
+ #define DEFINES_H
  #include "defines.h"
 #endif
 
 int unix_setup(void *gpw);
 int unix_crypt(void *given_key,void *gtext, int textlen, int secureflg); /* 0 or 1 */
-int shuffle(char deck[]);
\ No newline at end of file
+int shuffle(char deck[]);
