@@ -1,5 +1,5 @@
-#ifndef DEFINES_H_INC
- #define DEFINES_H_INC
+#ifndef DEFINES_H
+ #define DEFINES_H
  #include "defines.h"
 #endif
 #include "extra.h"
