@@ -33,6 +33,7 @@
 #include "lcrypt.h"
 #include "sunlink.h"
 #include "crc32.h"
+#include "locks.h"
 
  unsigned int bit7_flag=BITS7;
  unsigned int stream_flag=FALSE;
@@ -41,34 +42,37 @@
  int quiet=TRUE; /* silent by default */
  int unlink_flag=FALSE;
  int bare_flag=FALSE;
+ char *outfile=0;
+
 
 void usage(char *name)
 {
 
-     fprintf(stderr,_("Usage: %s (-e) FILE1 FILE2 ...\n"),name);
-     fprintf(stderr,_("Usage: %s -d FILE1 FILE2 ...\n\n"),name);
+     fprintf(stderr,_("Usage: mcrypt FILE1 FILE2 ...\n"));
+     fprintf(stderr,_("Usage: mcrypt -d FILE1 FILE2 ...\n\n"));
 
-     fprintf(stderr,_("\n  mini-crypt encrypts and decrypts files using the des, triple des \n\
+     fprintf(stderr,_("mini-crypt encrypts and decrypts files using the des, triple des \n\
 3-way, gost and blowfish algorithms.\n\
 \n\
- -e --encrypt      encrypts\n\
- -d --decrypt      decrypts\n\
- -t --type TYPE    specify the encryption and decryption algorithm.\n\
-   Legal parameters are: 3way,des,3des,gost,crypt and blowfish.\n\
-    (Default is blowfish)\n\
- -m --mode MODE    specify the mode (ecb or cbc). (Default is cbc)\n\
- -k --key  KEYWORD keyword to use.\n\
- -7 --7bit         keyword in 7bit mode. (default)\n\
- -8 --8bit         keyword in 8bit mode.\n\
- -u --unlink       unlink (and destroy) the input file after encryption/decryption\n\
- -L --license      displays license information.\n\
- -q --quiet        Suppresses some non critical warnings.\n\
- -V --verbose      Some more information is displayed.\n\
- -b --bare         Do not keep algorithm information in the encrypted file.\n\
- -f --force        forces output to stdout.\n\
+ -d --decrypt                 decrypts\n\
+ -a --algorithm=ALGORITHM     specify the encryption and decryption algorithm.\n\
+                              Legal parameters are: 3way, des, 3des, gost,\n\
+                              safer64, safer128, crypt and blowfish. \n\
+                              (Default is blowfish)\n\
+ -m --mode=MODE               specify the mode (ecb or cbc). (Default is cbc)\n\
+ -k --key=KEYWORD             keyword to use.\n\
+ -7 --7bit                    keyword in 7bit mode. (default)\n\
+ -8 --8bit                    keyword in 8bit mode.\n\
+ -u --unlink                  unlink (and destroy) the input file after\n\
+                              encryption/decryption\n\
+ -L --license                 displays license information.\n\
+ -q --quiet                   suppress some non critical warnings.\n\
+ -V --verbose                 some more information is displayed.\n\
+ -b --bare                    Do not keep algorithm information in the encrypted file.\n\
+ -f --force                   forces output to stdout.\n\
  \n\
- -h --help         prints this help\n\
- -v --version      prints the version number\n\
+ -h --help                    prints this help\n\
+ -v --version                 prints the version number\n\
  \n\
 Report bugs to nmav@i-net.paiko.gr.\n\n"));
 }
@@ -80,7 +84,7 @@ Report bugs to nmav@i-net.paiko.gr.\n\n"));
 int main(int argc,char **argv)
 {
 	short int ch,ein=0,din=0,kin=0,force=0,return_val=0;
-	char *einfile=0, *outfile=0 ,*dinfile=0, *keyword=0,*modet=0,*typet=0;
+	char *einfile=0 ,*dinfile=0, *keyword=0,*modet=0,*typet=0;
 	char **file; 
 	extern char *optarg;
 	int x,y,i,file_count=0;
@@ -90,10 +94,9 @@ struct option longopts[] = {
 	    { "version", 0, 0, 'v' },
 	    { "help", 0, 0, 'h' },
 	    { "bare", 0, 0, 'b' },	    
-	    { "encrypt", 0, 0, 'e' },
 	    { "decrypt", 0, 0, 'd' },
 	    { "mode", 1, 0, 'm' },
-	    { "type", 1, 0, 't' },
+	    { "algorithm", 1, 0, 'a' },
 	    { "key", 1, 0, 'k' },
 	    { "force", 0, 0, 'f' },
 	    { "7bit", 0, 0, '7' },	    
@@ -104,7 +107,18 @@ struct option longopts[] = {
 	    { "unlink", 0, 0, 'u'},	    
 	    { 0, 0, 0, 0 }
        	  };
- 
+
+#ifdef HAVE_SIGNAL_H
+ Signal( SIGINT, shandler);
+ Signal( SIGQUIT, shandler);
+ Signal( SIGSEGV, shandler);
+ Signal( SIGPIPE, shandler);   
+ Signal( SIGTERM, shandler);   
+ Signal( SIGHUP, shandler);   
+ Signal( SIGUSR1, SIG_IGN);
+ Signal( SIGUSR2, SIG_IGN);
+ Signal( SIGALRM, SIG_IGN);
+#endif 
 
 #ifdef ENABLE_NLS
   setlocale (LC_MESSAGES, "");
@@ -112,19 +126,27 @@ struct option longopts[] = {
   textdomain (PACKAGE);
 #endif
 
+/* Examine how we were called */
+   if (strstr(argv[0],"decrypt")!=NULL) {
+     strcpy( argv[0], "mdecrypt");
+     din=TRUE;
+     ein=FALSE;
+   }
+   else {
+     strcpy( argv[0], "mcrypt");
+     din=FALSE;
+     ein=TRUE; /* It will change by the arguments */
+   }
+
+
             /* File pointers are as much as the file arguments */
 
         file=malloc((argc-optind)*sizeof(char*)); 
 
-	while( (ch=getopt_long(argc,argv,"f78bLVuqedt:m:k:hv",longopts, (int *)0 )) != EOF )
+	while( (ch=getopt_long(argc,argv,"f78bLVuqda:m:k:hv",longopts, (int *)0 )) != EOF )
 
 		switch(ch) {
 	
-		case 'e':
-			ein=1;
-			if (din!=0) {fprintf (stderr, _("%s: You cannot specify both encryption and decryption.\n"),argv[0]);return 1;}
-			break;
-
 		case 'v':
 			printf (_("Mini-crypt v.%s\n"),VERSION);
 			printf (_("Copyright (C) 1998,1999 Nikos Mavroyanopoulos (nmav@i-net.paiko.gr)\n"));
@@ -151,12 +173,12 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
 			break;			
 
 		case 'd':
-			din=1;
-			if (ein!=0) {fprintf (stderr, _("%s: You cannot specify both encryption and decryption.\n"),argv[0]);return 1;}
+			din=TRUE;
+			ein=FALSE;
 			break;
 
 		case 'f':
-			force=1;
+			force=TRUE;
 			break;
 
 		case 'b':
@@ -194,7 +216,7 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
 			if (modet[0]=='e' || modet[0]=='E') mode=1;
 			free(modet);
 			break;
-		case 't':
+		case 'a':
 			typet= malloc(strlen(optarg)+1);
 			strcpy(typet,optarg);
 			if (strncasecmp(typet,"DES",3)==0) {
@@ -220,6 +242,16 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
                               if (strncasecmp(typet,"CRYPT",5)==0) {
                               type=CRYPT;
                               }
+                              else {
+                                if (strncasecmp(typet,"SAFER64",7)==0) {
+                                  type=SAFER64;
+                                }
+                                else {
+                                  if (strncasecmp(typet,"SAFER128",8)==0) {
+                                    type=SAFER128;
+                                  }
+                                }
+                              }
                              }
                            }
                           }
@@ -239,14 +271,14 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
 
 		}
 
-#ifdef HAVE_SYS_STAT_H
+#ifdef HAVE_UMASK
   umask(066);
 #endif
 
 /* Normal startup */
 
 
-
+/* ok now how many files were specified? */
     i=0;
     file_count+= (argc - optind);
 
@@ -260,13 +292,22 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"));
         i++;
     }
 
+    
     if (stream_flag==TRUE) file_count++;
 
 
 /* Do as many times as many files we got */
 for(i=0;i<file_count;i++) {
 
-    if(i!=0) { free(outfile); fprintf(stderr,"\n"); }
+    if(i!=0) { if (outfile!=NULL) free(outfile); }
+
+#ifdef HAVE_STAT
+    if ( is_normal_file( file[i]) == FALSE && stream_flag==FALSE ) {
+      fprintf( stderr, _("%s: %s is not a regular file. Skipping...\n"), argv[0], file[i]);
+      outfile=NULL;
+      continue; /* next */
+    }
+#endif
 
     if (stream_flag==FALSE) {
       if (quiet==FALSE) fprintf(stderr, _("File: %s\n"),file[i]);
@@ -275,29 +316,28 @@ for(i=0;i<file_count;i++) {
       if (quiet==FALSE) fprintf(stderr, _("File: stdin\n"));    
     }
     
-	if ( (strstr(argv[0],"decrypt")!=NULL) || (din==1)) {
+         /* Check how we were called */
+	if (din==TRUE) {
             if (stream_flag!=TRUE) dinfile=file[i];
-            din=1;
-             if( (isatty(fileno( (FILE *)(stdin) ) ) == 1) && (stream_flag==TRUE) && (force==0) ) { /* not a tty */        
-             fprintf(stderr,_("%s: Data will not be read from a terminal.\n"),argv[0]);
-             fprintf(stderr,_("Redirect the input instead.\nUse the --help parameter for more help.\n"));
-             exit(-1);
-             }
-        }
+            if ( (isatty(fileno( (FILE *)(stdin) ) ) == 1) && (stream_flag==TRUE) && (force==0) ) { /* not a tty */        
+              fprintf(stderr,_("%s: Encrypted data will not be read from a terminal.\n"),argv[0]);
+              fprintf(stderr,_("Redirect the input instead.\nUse the --help parameter for more help.\n"));
+               exit(-1);
+            }
+        } 
         else { /* called as mcrypt or something like that */
             if (stream_flag!=TRUE) einfile=file[i];
-            ein=1;
-             if( (isatty(fileno( (FILE *)(stdout) ) ) == 1) && (stream_flag==TRUE) && (force==0) ) { /* not a tty */        
-             fprintf(stderr,_("%s: Data will not be written to a terminal.\n"),argv[0]);
-             fprintf(stderr,_("Redirect the output instead.\nUse the --help parameter for more help.\n"));
-             exit(-1);
-             }            
+            if ( (isatty(fileno( (FILE *)(stdout) ) ) == 1) && (stream_flag==TRUE) && (force==0) ) { /* not a tty */        
+              fprintf(stderr,_("%s: Encrypted data will not be written to a terminal.\n"),argv[0]);
+              fprintf(stderr,_("Redirect the output instead.\nUse the --help parameter for more help.\n"));
+              exit(-1);
+            }            
         }
 
 
-
+        /* If no streams make the extensions */
        if(stream_flag!=TRUE) {
-	 if (din==1) {
+	 if (din==TRUE) {
 
             y=strlen(dinfile);
             /* If the file finished in .enc remove it */
@@ -306,9 +346,9 @@ for(i=0;i<file_count;i++) {
 
 	      outfile= calloc(y-3,1);
   	      strncat(outfile,dinfile,y-4);
-#ifdef HAVE_SYS_STAT_H
+#ifdef HAVE_STAT
               /* But if it exists exit */
-              if (check_file(outfile)!=0) {fprintf(stderr,_("\nFile %s exists... skipping...\n\n"),outfile); fflush(stderr); continue;}
+              if (check_file(outfile)!=0) {fprintf(stderr,_("%s: file %s exists... skipping...\n"), argv[0], outfile); continue;}
 #endif  	      
             }
             else { /* append .dec */
@@ -321,7 +361,15 @@ for(i=0;i<file_count;i++) {
 	 else { /* encryption- append .enc */
 	    outfile= calloc(strlen(einfile)+5,1);
 	    strcpy(outfile,einfile);
+            /* if file has already the .enc ignore it */
+            if (strstr( outfile, ".enc")!=NULL) {fprintf(stderr,_("%s: file %s has the .enc suffix... skipping...\n"), argv[0], outfile); continue;}
+
 	    strcat(outfile,".enc");
+#ifdef HAVE_STAT
+              /* But if it exists exit */
+              if (check_file(outfile)!=0) {fprintf(stderr,_("%s: file %s exists... skipping...\n"), argv[0], outfile); continue;}
+#endif  	      
+
 	 }
        }
        else { /* if streams */
@@ -336,7 +384,7 @@ for(i=0;i<file_count;i++) {
 
            if (bare_flag==FALSE) {
              if (check_file_head(file[i],&type,&mode, &bit7_flag, stream_flag)!=FALSE) {
-               fprintf(stderr,_("No headers. Assuming encrypted file with Unix crypt.\n")); 
+               fprintf(stderr,_("%s: no headers. Assuming encrypted file with Unix crypt.\n"),argv[0]); 
                type=CRYPT;
              }
            }
@@ -347,6 +395,8 @@ for(i=0;i<file_count;i++) {
 			case ThreeWAY:
 			case BLOWFISH:
 			case GOST:
+			case SAFER64:
+			case SAFER128:
 			 x=decrypt_general(type,dinfile,outfile,keyword);
 			 break;
 			case CRYPT:
@@ -364,18 +414,17 @@ for(i=0;i<file_count;i++) {
         	            else {
         	              if (quiet==FALSE) fprintf (stderr,_("Stdin was decrypted.\n"));
         	            }
-        	           
-#ifdef HAVE_SYS_STAT_H
+#ifdef HAVE_STAT        	           
 # ifdef HAVE_UTIME_H
                            if (stream_flag==FALSE) copyDate(dinfile,outfile);
-#endif
-        	           if (unlink_flag==TRUE && stream_flag==FALSE) x=safe_unlink(dinfile);
-	        	   if (x!=0) perror("safe_unlink");
+# endif
+        	           if (unlink_flag==TRUE && stream_flag==FALSE) remove(dinfile);
 #endif        	           
         	        }
 	        	else {
                           if (stream_flag==FALSE) {
         	           fprintf (stderr,_("File %s was NOT decrypted successfully.\n"),dinfile);
+          	           remove(outfile);
         	          }
         	          else {
         	           fprintf (stderr,_("Stdin was NOT decrypted successfully.\n"));
@@ -386,13 +435,15 @@ for(i=0;i<file_count;i++) {
 	}		
 
 /* Encrypt */
-	if (ein==1) {
+	if (ein==TRUE) {
 		      switch (type) {
 		        case DES:
 		        case TripleDES:
 	        	case ThreeWAY:
 	        	case BLOWFISH:
 	        	case GOST:
+	        	case SAFER64:
+	        	case SAFER128:
 	        	 x=encrypt_general(type,einfile,outfile,keyword);
 	        	 break;
 	        	 
@@ -412,21 +463,23 @@ for(i=0;i<file_count;i++) {
                           else {
 	        	   if (quiet==FALSE) fprintf (stderr,_("Stdin was successfully encrypted.\n"));
 	        	  }
-#ifdef HAVE_SYS_STAT_H
+#ifdef HAVE_STAT
 # ifdef HAVE_UTIME_H
                            if (stream_flag==FALSE) copyDate(einfile,outfile);
 # endif
+#endif
 	        	   if (unlink_flag==TRUE && stream_flag==FALSE) x=safe_unlink(einfile);
 	        	   if (x!=0) perror("safe_unlink");
-#endif	        	   
 		       	}
 	                else {
                          if (stream_flag==FALSE) {
 	                  fprintf (stderr,_("File %s was NOT encrypted successfully.\n"),einfile);
+                          remove(outfile);
 	                 }
                          else {
 	                  fprintf (stderr,_("Stdin was NOT encrypted successfully.\n"));
 	                 }
+
 	                }
 	                
 	                return_val+=x;
@@ -452,7 +505,10 @@ char *tmp;
 FILE *TOF;
 FILE *FROMF;
 
-/* Crypt does not care about these */
+/* Crypt does not care about these 
+ * It's purpose is to be compatible with the unix crypt not safe.
+ * It isn't a block algorithm after all.
+ */
 mode=2;
 bit7_flag=BITS8;
 
@@ -464,24 +520,28 @@ if (stream_flag==TRUE) {
 }
 else {
 
-	FROMF=fopen(fromfile,"r");
+	FROMF=fopen(fromfile,"rb");
 	if (FROMF==NULL) {perror("fopen");return 1;}
-	TOF=fopen(tofile,"w");
+	read_lock( fileno((FILE*)FROMF));
+	
+	TOF=fopen(tofile,"wb");
 	if (TOF==NULL) {perror("fopen");return 1;}
+	write_lock( fileno((FILE*)TOF));
 }
 
-        bzero(keyword,sizeof(keyword));
+        Bzero(keyword,sizeof(keyword));
 
 	if (key==NULL) 	{
+          if (quiet==TRUE && stream_flag==FALSE) fprintf(stderr, _("File: %s\n"),fromfile);
           tmp=get_password(8,bit7_flag,DECRYPT,type);
           if (tmp==NULL) return 1;
-
-          memcpy(keyword,tmp,strlen(tmp));
+          memmove(keyword, tmp, strlen(tmp));
+          Bzero(tmp, strlen(tmp));
 
 	}
 	else {
-           if (strlen(key)>8) {key[8]='\0'; fprintf(stderr,_("Warning: Reducing the size of the key\n"));}
-	   memcpy(keyword,key,strlen(key)); /* Copy the 64 bits in key */
+           if (strlen(key) > 8) {key[8]='\0'; fprintf(stderr,_("Warning: Reducing the size of the key\n"));}
+	   memmove(keyword, key, strlen(key)); /* Copy the 64 bits in key */
 	}
 
 
@@ -495,7 +555,7 @@ do
 /* else read the first 1 byte of the file and stores the to plaintext */
  
                 how=fread(plaintext, 1, sizeof(plaintext), FROMF);
-                if (how==0) {
+                if (how == 0 ) {
                  if (feof(FROMF)!=0) {
                   ch=1;
                  }
@@ -514,9 +574,17 @@ do
 while(ch==0);
 
 /* close files */
+ if (stream_flag==FALSE) {
+   fflush(FROMF);
+   fflush(TOF);
+   unlock( fileno((FILE*)TOF));
+   unlock( fileno((FILE*)FROMF));
+ }
    fclose(TOF);
    fclose(FROMF);
-
+   
+   Bzero(keyword, sizeof(keyword));
+   
 return 0;
 }
 
@@ -545,24 +613,30 @@ FILE *RTOF;
 FILE *FROMF;
 int keysize=get_key_size(algorithm);
 int blocksize=get_block_size(algorithm);
+char tmpkey[128];
 
-word32 *IV=malloc(blocksize);
+word32 *IV=0;
 word32 *keyword=calloc(keysize,1); 
 word32 *ciphertext=malloc(blocksize);
-word32 *ciphertextcbc=malloc(blocksize);
+word32 *ciphertextcbc=0;
 FILE * TOF; 
 
  TOF=tmpfile();
- 
+ write_lock( fileno((FILE*)TOF));
+        
+
+if (mode==CBC) {
+ IV=malloc(blocksize);
+ ciphertextcbc=malloc(blocksize); 
  /* Initialisation Vector */
  srand(time(0));
 
 /* Random number generator */
 
- for (i=0;i<blocksize/4;i++) {
+ for (i=0;i<blocksize/sizeof(word32);i++) {
   IV[i] = RAND32 ;
  }
-
+}
 
 /* open files */
   if (stream_flag==TRUE) {
@@ -570,28 +644,34 @@ FILE * TOF;
 	RTOF=stdout;
   }
   else {
-	FROMF=fopen(fromfile,"r");
+	FROMF=fopen(fromfile,"rb");
 	if (FROMF==NULL) {perror("fopen");return 1;}
-
-	RTOF=fopen(tofile,"w");
+        read_lock( fileno((FILE*)FROMF));
+        
+	RTOF=fopen(tofile,"wb");
 	if (RTOF==NULL) {perror("fopen");return 1;}
+        write_lock( fileno((FILE*)RTOF));
+        
   }
 
 
 
   if (bare_flag==FALSE) {
     write_file_head( fileno((FILE*)TOF), type, mode, bit7_flag);
-  }
-//  fseek(TOF, blocksize, SEEK_CUR); /* move blocksize bytes */
-write(fileno((FILE*)TOF),ciphertext,blocksize);
+    write( fileno((FILE*)TOF), ciphertext, blocksize);
+     /* Just leave space to put later the crc32 */
+  }  
+
 
 	if (key==NULL) 	{
+          if (quiet==TRUE && stream_flag==FALSE) fprintf(stderr, _("File: %s\n"),fromfile);
           tmp=get_password( keysize, bit7_flag, ENCRYPT, algorithm );
           if (tmp==NULL) return 1;
           len=strlen(tmp);
 
           if (len > keysize) len=keysize;
-          memcpy(keyword, tmp, len);
+          memmove(keyword, tmp, len);
+          Bzero(tmp, len);
                 
 	}
 	else {
@@ -603,20 +683,22 @@ write(fileno((FILE*)TOF),ciphertext,blocksize);
                   fprintf(stderr,_("Warning: Reducing the size of the key\n"));
                   len=keysize;
                }
-               memcpy(keyword, key, len);
+               memmove(keyword, key, len);
            }
            else { /* 7bit mode */
-               strip8bit(key, strlen(key));
                len=strlen(key);
+               memmove(tmpkey, key, len);
+               strip8bit(tmpkey, len);
+               len=strlen(tmpkey);
                if (len > keysize) {
                   fprintf(stderr,_("Warning: Reducing the size of the key\n"));
                   len=keysize;
                }
-               memcpy(keyword, key, len);
+               memmove(keyword, tmpkey, len);
+               Bzero(tmpkey, len); /* key is not bzero'd because it may be used again */
            }           
 	}
 
-
  if (quiet==FALSE) show_mode(ENCRYPT,algorithm,mode,tofile,bit7_flag);
 
  if (algorithm != BLOWFISH) {
@@ -635,13 +717,15 @@ write(fileno((FILE*)TOF),ciphertext,blocksize);
 
 /* Encryption Starts here */
 
-if (mode==CBC) {
+switch (mode) {
+
+ case CBC:
   do
   {
 
     /* If first time use the Init. vector */
       if (start==0) {
-        memcpy(ciphertext,IV,blocksize);
+        memmove(ciphertext,IV,blocksize);
         start=1;
       }
       else {
@@ -664,7 +748,7 @@ if (mode==CBC) {
         if (bare_flag==FALSE) crc32(ciphertext, how);
 
        /* for cbc xor the plaintext with the old ciphertext block */
-          for(j=0; j<blocksize/4; j++)
+          for(j=0; j<blocksize/sizeof(word32); j++)
             ciphertext[j] ^= ciphertextcbc[j];
 
       }
@@ -673,25 +757,24 @@ if (mode==CBC) {
      mcrypt(algorithm, keyword, len, ciphertext);
 
      /* for cbc keep the old ciphertext */
-     memcpy(ciphertextcbc,ciphertext,blocksize);
+     memmove(ciphertextcbc,ciphertext,blocksize);
 
          
      if (fwrite(ciphertext, 1, blocksize, TOF) == 0) {perror("fwrite"); return 1;}
       if (ch==1) if (fwrite(&counter, 1, sizeof(counter), TOF) == 0) {perror("fwrite"); return 1;}
   }
   while(ch==0);
+  break;
 
-}
-else { /* mode==ECB */
-
+ case ECB:
   do
   {
 
-  /* else read the first n bytes of the file and store to ciphertext */
+  /* read the first n bytes of the file and store to ciphertext */
  
                    how=fread(ciphertext,1,blocksize,FROMF);
 
-  	           if (how<blocksize) {
+  	           if (how < blocksize) {
                     if ( how ==0 && feof(FROMF)==0) {
                        perror("fread"); 
                        return 1;
@@ -716,8 +799,9 @@ else { /* mode==ECB */
 
   }
   while(ch==0);
+  break;
 
-} /* if mode=xy */
+} /*  switch */
 
  if (bare_flag==FALSE) {
     /* Put the crc32 in byte 6 */
@@ -730,9 +814,9 @@ else { /* mode==ECB */
 
 /* encrypt the crc32, but first put some random data there 
  */
-   for(j=0; j<blocksize/4; j++)
+   for(j=0; j<blocksize/sizeof(word32); j++)
      ciphertext[j] = RAND32;
-   memcpy(ciphertext,&fcrc32,sizeof(fcrc32));
+   memmove(ciphertext,&fcrc32,sizeof(fcrc32));
    mcrypt(algorithm, keyword, len, ciphertext);
    fseek(TOF,6,SEEK_SET);
    if (fwrite(ciphertext, 1, blocksize, TOF) == 0) {perror("fwrite"); return 1;}
@@ -752,18 +836,33 @@ else { /* mode==ECB */
  free(copychar);
 /* End of copy */
    
-/* close files */
-    fclose(TOF);
+/* close and unlock files*/
 
+    fflush(TOF);
+    unlock( fileno((FILE*)TOF) );
+    
+ if (stream_flag==FALSE) {
+    fflush(FROMF);
+    fflush(RTOF);    
+    unlock( fileno((FILE*)FROMF) );
+    unlock( fileno((FILE*)RTOF) );
+ }
+                
+    fclose(TOF);
     end_mcrypt(algorithm);
   
     fclose(RTOF);
     fclose(FROMF);
+/* Ready */
 
-    free(IV);
+    if (mode==CBC) {
+     free(IV);
+     free(ciphertextcbc);     
+    }
+
+    Bzero(keyword, keysize);
     free(keyword); 
     free(ciphertext);
-    free(ciphertextcbc);
 
 
   return 0;
@@ -783,6 +882,7 @@ int i=0,j=0,ch=0,start=0;
 char *tmp;
 int keysize=get_key_size(algorithm);
 int blocksize=get_block_size(algorithm);
+char tmpkey[128];
 
 word32 *ciphertext_old=malloc(blocksize);
 word32 *buffer=malloc(blocksize);
@@ -803,10 +903,13 @@ if (stream_flag==TRUE) {
 }
 else {
 
-	FROMF=fopen(fromfile,"r");
+	FROMF=fopen(fromfile,"rb");
 	if (FROMF==NULL) {perror("fopen");return 1;}
-	RTOF=fopen(tofile,"w");
+        read_lock( fileno((FILE*)FROMF));
+        
+	RTOF=fopen(tofile,"wb");
 	if (RTOF==NULL) {perror("fopen");return 1;}
+        write_lock( fileno((FILE*)RTOF));
 
         if (bare_flag==FALSE) fseek( FROMF, 6, SEEK_CUR);
         /* In streams this has been done when checking for headers. */
@@ -815,12 +918,14 @@ else {
 
   
 	if (key==NULL) {
+          if (quiet==TRUE && stream_flag==FALSE) fprintf(stderr, _("File: %s\n"),fromfile);
           tmp=get_password(keysize,bit7_flag,DECRYPT,algorithm);
           if (tmp==NULL) return 1;
           len=strlen(tmp);
                               
           if (len>keysize) len=keysize;
-          memcpy(keyword, tmp, len);
+          memmove(keyword, tmp, len);
+          Bzero(tmp, len);
 
 	}
 	else {
@@ -830,16 +935,19 @@ else {
                  fprintf(stderr,_("Warning: Reducing the size of the key\n"));
                  len=keysize;
               }
-              memcpy(keyword, key, len);
+              memmove(keyword, key, len);
             }
             else { /* 7bit mode */
-               strip8bit(key,strlen(key));
                len=strlen(key);
+               memmove(tmpkey, key, len);
+               strip8bit( tmpkey, len);
+               len=strlen(tmpkey);
                if (len > keysize) {
                   fprintf(stderr,_("Warning: Reducing the size of the key\n"));
                   len=keysize;
                }
-               memcpy(keyword, key, len);
+               memmove(keyword, tmpkey, len);
+               Bzero(tmpkey, len);
             }          
         }
 
@@ -858,63 +966,67 @@ if (quiet==FALSE) show_mode(DECRYPT,algorithm,mode,tofile,bit7_flag);
  
  init_mcrypt(algorithm,keyword,len);
 
- if (bare_flag==FALSE) clear_crc32();
-
 /* Decrypt CRC32 stored in file */
  if (bare_flag==FALSE) {
 
+   clear_crc32();
+   
     /* Read CRC32 */
    fread(ciphertext, 1, blocksize, FROMF);
    mdecrypt(algorithm, keyword, len ,ciphertext);
-   memcpy(&fcrc32, ciphertext, sizeof(fcrc32));
+   memmove(&fcrc32, ciphertext, sizeof(fcrc32));
 #ifndef WORDS_BIGENDIAN
    fcrc32=byteswap(fcrc32);
 #endif
 
  }
 
+switch (mode) {
 
-/* decryption starts here */
- do
- {
+ case CBC:
+  /* decryption starts here */
+   do
+   {
 
-  /* reads the first n bytes of the file and stores the to ciphertext */
-  /* /4 is here because word32 is 4 bytes */
-  
-        for (i=0;i<blocksize/4;i++) { 
-           if(ch!=0) break;
-           if ((fread(&ciphertext[i], 1, sizeof(ciphertext[i]),FROMF)==0) && (feof(FROMF)==0)) {perror("fread"); return 1;}
-            
-           if (feof(FROMF)!=0) {
-            ch=1;
-            memcpy(&counter, &ciphertext[i], sizeof(counter));
-            if (counter >= blocksize) {fprintf(stderr,_("Incorrect file type\n")); return 1;}
-            clearerr(FROMF);
+    /* reads the first n bytes of the file and stores the to ciphertext */
+    /* /4 is here because word32 is 4 bytes */
+           i=fread( ciphertext, 1, blocksize, FROMF);
+
+           if (i < blocksize) {
+             if ( ferror(FROMF)!=0 ) {
+               perror("fread"); 
+               return 1;
+             }
+             else { /* End of file */
+               ch=1;
+               if (i==0) {fprintf(stderr,_("Incorrect file type\n")); return 1;}
+               memmove(&counter, ciphertext, sizeof(counter));
+               if (counter >= blocksize) {fprintf(stderr,_("Incorrect file type\n")); return 1;}
+             }
            }
-         }
 
          /* Decrypt the ciphertext */
 
-        memcpy(ciphertext_orig,ciphertext,blocksize); /* keep backup */
+        memmove(ciphertext_orig,ciphertext,blocksize); /* keep backup */
 
 
         mdecrypt(algorithm, keyword, len, ciphertext);
 	
-/* if cbc: xor the new plaintext with the old ciphertext */
-        if (mode==CBC && start>0) {
-          for(j=0; j<blocksize/4; j++) {
+  /* if cbc: xor the new plaintext with the old ciphertext */
+        if (start>0) {
+          for(j=0; j<blocksize/sizeof(word32); j++) {
             ciphertext[j] ^= ciphertext_old[j];
           }
         }
 
-        memcpy(ciphertext_old,ciphertext_orig,blocksize);
+        memmove(ciphertext_old,ciphertext_orig,blocksize);
 
 
 /* Buffer isn't immediately written here. It waits a round to do
  * so we can keep track of the end of file garbage and the IV at the begining
  */
 
-         if (start>1 || mode==ECB) {
+         if (start>1) {
  
 		if (ch==1) {
                   if (counter==0) break;
@@ -924,50 +1036,124 @@ if (quiet==FALSE) show_mode(DECRYPT,algorithm,mode,tofile,bit7_flag);
 
 		}
 		else {
-		    if (mode!=1 || start!=0) {
+
+   		      if (fwrite(buffer,blocksize,1,RTOF)==0) {perror("fwrite"); return 1;}
+                       /* CRC32 */
+                      if (bare_flag==FALSE) crc32(buffer, blocksize);
+
+                }
+         }
+
+
+
+         memmove(buffer,ciphertext,blocksize);  /* copy the plaintext to buffer */
+
+         start++; /* ok enter that functions */
+
+   }
+   while(ch==0);
+   break;
+
+
+ case ECB:
+  /* decryption starts here */
+   do
+   {
+
+    /* reads the first n bytes of the file and stores the to ciphertext */
+
+           i=fread( ciphertext, 1, blocksize, FROMF);
+
+           if (i < blocksize) {
+             if (ferror(FROMF)!=0) {
+               perror("fread"); 
+               return 1;
+             }
+             else { /* eof */
+               ch=1;
+               if (i==0) {fprintf(stderr,_("Incorrect file type\n")); return 1;}
+               memmove(&counter, ciphertext, sizeof(counter));
+               if (counter >= blocksize) {fprintf(stderr,_("Incorrect file type\n")); return 1;}
+             }
+           }
+
+         /* Decrypt the ciphertext */
+
+        memmove(ciphertext_orig,ciphertext,blocksize); /* keep backup */
+
+
+        mdecrypt(algorithm, keyword, len, ciphertext);
+	
+        memmove(ciphertext_old,ciphertext_orig,blocksize);
+
+
+/* Buffer isn't immediately written here. It waits a round to do
+ * so we can keep track of the end of file garbage and the IV at the begining
+ */
+
+		if (ch==1) {
+                  if (counter==0) break;
+		  if (fwrite(buffer,(int)counter,1,RTOF)==0) {perror("fwrite"); return 1;}
+  		   /* CRC32 */
+          	  if (bare_flag==FALSE) crc32(buffer, (int)counter);
+
+		}
+		else {
+		    if (start!=0) {
    		      if (fwrite(buffer,blocksize,1,RTOF)==0) {perror("fwrite"); return 1;}
                        /* CRC32 */
                       if (bare_flag==FALSE) crc32(buffer, blocksize);
 
    		    }
                 }
-         }
+
+         start++;
+         memmove(buffer,ciphertext,blocksize);  /* copy the plaintext to buffer */
+
+   }
+   while(ch==0);
+   break;
+} /* switch */
 
 
+/* close and unlock files */
 
-         memcpy(buffer,ciphertext,blocksize);  /* copy the plaintext to buffer */
-
-         start++; /* ok enter that functions */
-
+ if (stream_flag==FALSE) {
+  fflush(FROMF);
+  fflush(RTOF);
+  unlock( fileno((FILE*)FROMF));
+  unlock( fileno((FILE*)RTOF));
  }
- while(ch==0);
+ 
+  fclose(RTOF);
+  fclose(FROMF);
 
-/* close files */
-    fclose(RTOF);
-    fclose(FROMF);
+/* ready */
 
-    end_mcrypt(algorithm);  
+  end_mcrypt(algorithm);  
 
   free(ciphertext_old);
   free(buffer);
   free(ciphertext);
   free(ciphertext_orig);
+
+  Bzero(keyword, keysize);
   free(keyword);
 
  if (bare_flag==FALSE) {
 
-  newcrc32=get_crc32();
-
-  if (newcrc32==fcrc32) {
-   return 0; /* crc check passed */
-  }
-  else {
-   fprintf(stderr, _("CRC32 check failed\n"));
-   return 1; /* CRC32s Do not match */
-  }
+    newcrc32=get_crc32();
+    if (newcrc32==fcrc32) {
+     return 0; /* crc check passed */
+    }
+    else {
+     fprintf(stderr, _("CRC32 check failed\n"));
+     return 1; /* CRC32s Do not match */
+    }
  }
-else { /* no crc is stored here */
- return 0;
-}
+ else { /* no crc is stored here */
+   return 0;
+ }
+
 }
 
